/*
------------------------------------------------------------
	ce_backend.h - Backend "API"
------------------------------------------------------------
 * Copyright (C) David Olofson, 2002
 *
 * This code is released under the terms of the GNU LGPL.
 */

#ifndef _CE_BACKEND_H_
#define _CE_BACKEND_H_

#include <stdarg.h>

#include "conemu.h"


/*----------------------------------------------------------
	Rendering backend definition.
----------------------------------------------------------*/
typedef struct ce_backend_t
{
	/*------------------------------------------------------------
	   These *must* be implemented to make a valid backend!
	------------------------------------------------------------*/

	/* Open. Should set up _ce_screen_[wh] to indicate the
	 * actual size of the obtained screen in characters.
	 */
	int (*open)(int w, int h, int flags);

	/* Close... */
	void (*close)(void);

	/* Get the next ASCII character from the keyboard */
	int (*getc)(void);

	/* Render a block of characters to the screen */
	int (*put_fast)(ce_rect_t rect, ce_character_t *buffer, int pitch);


	/*------------------------------------------------------------
	   These have "working" default implementations that
	   you may rely on, or override if you have to, or to
	   optimize performance. Some represent functions that
	   will simply be reported as "not supported" if the
	   respective callback is missing. (Ex: kbhit())
	------------------------------------------------------------*/

	/* Quickly put a single character. Screen coords! */
	int (*put)(int x, int y, ce_character_t c);

	/* Return 1 if a key is available for reading */
	int (*kbhit)(void);

	/*
	 * Read and parse a line from the keyboard. Preferably,
	 * a line editor should be provided for this, and it
	 * should show up at the current output cursor position.
	 */
	int (*scanf)(ce_window_t *window, const char *fmt, va_list args);

	/*
	 * putc() puts 'c' at the current cursor output posn.
	 * puts() writes a string with control codes.
	 * printf() does the usual C formatting.
	 *
	 * putchar(), puts() and printf() should understand
	 * the following control codes:
	 *
	 *	'\t'	Pad with blanks to the next tab posn.
	 *	(9)	(And tabs *are* 8 characters, period! ;-)
	 *
	 *	'\n'	Return to left margin, move one line down.
	 *	(10)	Moving past the bottom line of the buffer
	 *		should scroll the whole buffer.
	 *
	 *	'\r'	Return to left margin only.
	 *	(13)
	 *
	 * NOTE:
	 *	If you implement any of these three, you will have
	 *	to manage buffer updates, cursor movement and
	 *	refreshing of the screen!
	 */
	int (*putchar)(ce_window_t *window, int c);
	int (*puts)(ce_window_t *window, const char *s);
	int (*printf)(ce_window_t *window, const char *fmt, va_list args);

	int (*redirect_stdio)(ce_window_t *window);

	/* Place the output cursor. */
	void (*locate)(ce_window_t *window, int x, int y);
	int (*where_x)(ce_window_t *window);
	int (*where_y)(ce_window_t *window);

	/* Place visible cursor at (x, y); *screen* coordinates */
	int (*cursor_at)(int x, int y);

	/*
	 * Set palette entry 'index' to the best match of
	 * the color defined by 'r', 'g' and 'b'.
	 */
	void (*set_palette)(unsigned char index,
		unsigned char r, unsigned char g, unsigned char b);
} ce_backend_t;


/*----------------------------------------------------------
	Internal variables (available to backends)
----------------------------------------------------------*/

/* Current screen size in characters */
extern int _ce_screen_w;
extern int _ce_screen_h;

/* Window list in Z order */
extern int _ce_max_windows;
extern int _ce_windows;
extern ce_window_t **_ce_wintab;

extern ce_window_t *_ce_root_window;
extern ce_window_t *_ce_cursor_window;

/* Currently used backend */
extern ce_backend_t *_ce_backend;

/*
 * Colors and attributes applied to characters
 * printed "through the cursor". (Does not affect
 * low level calls that use ce_character_t.)
 */
extern ce_character_t _ce_cursor;

#endif /*_CE_BACKEND_H_*/
