/*
------------------------------------------------------------
	ce_sdl.c - SDL backend for the Console Emulator
------------------------------------------------------------
 * Copyright (C) David Olofson, 2002
 *
 * This code is released under the terms of the GNU LGPL.
 */

#include <stdio.h>
#include "SDL.h"
#include "ce_sdl.h"

#define	FONT	"confont.bmp"
#define	FONTW	8
#define	FONTH	14

SDL_Surface *screen = NULL;
SDL_Surface *font = NULL;

Uint32	colormap[256];
Uint32	cursor_colors[16];

/*----------------------------------------------------------
	H/W Cursor emulation
----------------------------------------------------------*/
static int cursor_x = -1;
static int cursor_y = -1;
static Uint32 last_draw_cursor;

static void _draw_cursor();
static void _remove_cursor();


/*----------------------------------------------------------
	Backend Callbacks
----------------------------------------------------------*/

static void sdl_set_palette(unsigned char index,
		unsigned char r, unsigned char g, unsigned char b)
{
	colormap[index] = SDL_MapRGB(screen->format, r, g, b);
}


static int sdl_open(int w, int h, int flags)
{
	int i;
	SDL_Surface *tmp;
	int sflags = 0;
	if(flags & CE_OPEN_FULLSCREEN)
		sflags |= SDL_FULLSCREEN;

	if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_EVENTTHREAD) < 0)
	{
		fprintf(stderr, "conemu: Failed to initialize SDL!\n");
		return -1;
	}

	font = SDL_LoadBMP(FONT);
	if(!font)
	{
		fprintf(stderr, "conemu: Failed to load font!\n");
		SDL_Quit();
		return -1;
	}

	screen = SDL_SetVideoMode(w * FONTW, h * FONTH, 0, sflags);
	if(!screen)
	{
		fprintf(stderr, "conemu: Failed to set video mode!\n");
		SDL_FreeSurface(font);
		font = NULL;
		SDL_Quit();
		return -1;
	}

	SDL_SetColorKey(font, SDL_SRCCOLORKEY | SDL_RLEACCEL, 
			SDL_MapRGB(screen->format, 0, 0, 0));
	tmp = SDL_DisplayFormat(font);
	if(tmp)
	{
		SDL_FreeSurface(font);
		font = tmp;
	}
	else
		fprintf(stderr, "conemu: WARNING: Failed to convert font!\n");

	SDL_EnableUNICODE(1);
	SDL_EnableKeyRepeat(250, 25);

	for(i = 0; i < 8; ++i)
		cursor_colors[15-i] = SDL_MapRGB(screen->format,
				255, (i+8)<<4, (i+8)<<3);
	for(i = 0; i < 8; ++i)
		cursor_colors[7-i] = SDL_MapRGB(screen->format,
				i<<5, i<<4, i<<3);

	_ce_screen_w = w;
	_ce_screen_h = h;
	_ce_backend = &_ce_sdl;
	return 1;
}


static void sdl_close(void)
{
	SDL_FreeSurface(font);
	font = NULL;
	SDL_Quit();
}


static int sdl_getc(void)
{
	SDL_Event ev;
	while(1)
	{	
		/*
		 * Yeah, this sucks.
		 * Then again, if you care, just don't use ce_getc()
		 * without first checking with ce_kbhit()!
		 */
		while(!SDL_PollEvent(&ev))
		{
			_draw_cursor();
			SDL_Delay(10);
		}

		if(ev.type != SDL_KEYDOWN)
			continue;

		switch(ev.key.keysym.sym)
		{
		  case SDLK_LEFT:
			return CE_LEFT;
		  case SDLK_RIGHT:
			return CE_RIGHT;
		  case SDLK_UP:
			return CE_UP;
		  case SDLK_DOWN:
			return CE_DOWN;
		  case SDLK_DELETE:
			return CE_DEL;
		  case SDLK_BACKSPACE:
			return CE_BS;
		  default:
			if(ev.key.keysym.unicode & 0xff80)
				return 0;
			else
				return ev.key.keysym.unicode;
		}
	}
}


static int sdl_put(int x, int y, ce_character_t c)
{
	SDL_Rect from, to;
	from.w = FONTW;
	from.h = FONTH;
	from.x = c.c % 16 * FONTW;
	from.y = c.c / 16 * FONTH;
	to.x = x * FONTW;
	to.y = y * FONTH;
	to.w = FONTW;
	to.h = FONTH;
	SDL_FillRect(screen, &to, colormap[c.bg]);
	SDL_BlitSurface(font, &from, screen, &to);
	SDL_UpdateRects(screen, 1, &to);
	return 0;
}


static int sdl_put_fast(ce_rect_t rect, ce_character_t *buffer, int pitch)
{
	SDL_Rect r;
	int x, y;
	for(y = 0; y < rect.h; ++y)
		for(x = 0; x < rect.w; ++x)
		{
			SDL_Rect from, to;
			ce_character_t c = buffer[y*pitch + x];
			from.w = FONTW;
			from.h = FONTH;
			from.x = c.c % 16 * FONTW;
			from.y = c.c / 16 * FONTH;
			to.x = x * FONTW;
			to.y = y * FONTH;
			to.w = FONTW;
			to.h = FONTH;
			SDL_FillRect(screen, &to, colormap[c.bg]);
			SDL_BlitSurface(font, &from, screen, &to);
		}
	r.x = rect.x * FONTW;
	r.y = rect.y * FONTH;
	r.w = rect.w * FONTW;
	r.h = rect.h * FONTH;
	SDL_UpdateRects(screen, 1, &r);
	return 0;
}


static int sdl_kbhit(void)
{
	SDL_Event ev;
	/* In case someone's polling like wild... */
	if(SDL_GetTicks() - last_draw_cursor > 33)
		_draw_cursor();
	return (SDL_PeepEvents(&ev, 1, SDL_PEEKEVENT, SDL_KEYDOWNMASK) > 0);
}


static int sdl_cursor_at(int x, int y)
{
	_remove_cursor();
	cursor_x = x;
	cursor_y = y;
	_draw_cursor();
	return 0;
}


/*----------------------------------------------------------
	Backend Interface
----------------------------------------------------------*/
ce_backend_t _ce_sdl = {
	/* --- required --- */
	sdl_open,	/* open */
	sdl_close,	/* close */
	sdl_getc,	/* getc */
	sdl_put_fast,	/* put_fast */

	/* --- optional --- */
	sdl_put,	/* put */

	sdl_kbhit,	/* kbhit */
	NULL,		/* scanf */
	NULL,		/* putchar */
	NULL,		/* puts */
	NULL,		/* printf */

	NULL,		/* redirect_stdio */

	NULL,		/* locate */
	NULL,		/* where_x */
	NULL,		/* where_y */

	sdl_cursor_at,	/* cursor_at */
	sdl_set_palette	/* set_palette */
};


/*----------------------------------------------------------
	H/W Cursor emulation
----------------------------------------------------------*/
static void _draw_cursor()
{
	ce_rect_t r;
	ce_character_t c;
	SDL_Rect from, to;
	if(cursor_x < 0 || cursor_y < 0 ||
			cursor_x >= _ce_screen_w ||
			cursor_y >= _ce_screen_h)
		return;

	r.x = cursor_x;
	r.y = cursor_y;
	r.w = r.h = 1;
	ce_get_screen(r, &c, 1);

	from.w = FONTW;
	from.h = FONTH;
	from.x = c.c % 16 * FONTW;
	from.y = c.c / 16 * FONTH;
	to.x = cursor_x * FONTW;
	to.y = cursor_y * FONTH;
	to.w = FONTW;
	to.h = FONTH;
	SDL_FillRect(screen, &to, cursor_colors[(SDL_GetTicks() / 33) & 15]);
	SDL_BlitSurface(font, &from, screen, &to);
	SDL_UpdateRects(screen, 1, &to);

	last_draw_cursor = SDL_GetTicks();
}


static void _remove_cursor()
{
	ce_rect_t r;
	ce_character_t c;
	if(cursor_x < 0 || cursor_y < 0 ||
			cursor_x >= _ce_screen_w ||
			cursor_y >= _ce_screen_h)
		return;
	r.x = cursor_x;
	r.y = cursor_y;
	r.w = r.h = 1;
	ce_get_screen(r, &c, 1);
	sdl_put(cursor_x, cursor_y, c);
}
