/*
------------------------------------------------------------
	conemu.c - Text Console Emulator / Wrapper
------------------------------------------------------------
 * Copyright (C) David Olofson, 2002
 *
 * This code is released under the terms of the GNU LGPL.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "conemu.h"

#include "ce_backend.h"
#include "ce_sdl.h"
#include "ce_stdio.h"

#ifdef getc
#undef	getc
#endif


/*----------------------------------------------------------
	Internal variables (available to backends)
----------------------------------------------------------*/

int _ce_screen_w = 0;
int _ce_screen_h = 0;

int _ce_max_windows = 0;
int _ce_windows = 0;
ce_window_t **_ce_wintab = NULL;
ce_window_t *_ce_root_window = NULL;
ce_window_t *_ce_cursor_window = NULL;

ce_backend_t *_ce_backend = NULL;

ce_character_t *_ce_tempbuf = NULL;


/*----------------------------------------------------------
	Local stuff
----------------------------------------------------------*/

static ce_character_t _ce_blank = {' ', 0, 0, 0};

/*
 * "h/w" cursor position
 */
static int _ce_cursor_x = 0;
static int _ce_cursor_y = 0;

static unsigned char colors[16][3] = {
	{0x00,0x00,0x00},	/* CE_BLACK */
	{0x00,0x00,0x80},	/* CE_BLUE */
	{0x00,0x80,0x00},	/* CE_GREEN */
	{0x00,0x80,0x80},	/* CE_CYAN */
	{0x80,0x00,0x00},	/* CE_RED */
	{0x80,0x00,0x80},	/* CE_MAGENTA */
	{0x80,0x00,0x00},	/* CE_BROWN */
	{0xb0,0xb0,0xb0},	/* CE_LIGHTGRAY */

	{0x60,0x60,0x60},	/* CE_DARKGRAY */
	{0x00,0x00,0xff},	/* CE_LIGHTBLUE */
	{0x00,0xff,0x00},	/* CE_LIGHTGREEN */
	{0x00,0xff,0xff},	/* CE_LIGHTCYAN */
	{0xff,0x00,0x00},	/* CE_LIGHTRED */
	{0xff,0x00,0xff},	/* CE_LIGHTMAGENTA */
	{0xff,0xff,0x00},	/* CE_YELLOW */
	{0xff,0xff,0xff}	/* CE_WHITE */
};


static void load_palette(void)
{
	int i;

	/* Standard CGA/EGA colors */
	for(i = 0; i < 16; ++i)
		ce_set_palette(i, colors[i][0],  colors[i][1],  colors[i][2]);

	/* Various 4 bit gradients */
	for(i = 0; i < 16; ++i)
	{
		ce_set_palette(i+16, i<<4, 0, 0);	/* black -> red */
		ce_set_palette(i+32, 0, i<<4, 0);	/* black -> green */
		ce_set_palette(i+48, 0, 0, i<<4);	/* black -> blue */
		ce_set_palette(i+64, i<<4, i<<4, 0);	/* black -> yellow */
		ce_set_palette(i+80, i<<4, 0, i<<4);	/* black -> magenta */
		ce_set_palette(i+96, 0, i<<4, i<<4);	/* black -> cyan */
		ce_set_palette(i+112, i<<4, i<<4, i<<4);/* black -> white */
	}

	/* 7 bit black -> white gradient */
	for(i = 0; i < 128; ++i)
		ce_set_palette(i+128, i<<1, i<<1, i<<1);
}


/*
 * Inserts a window at z-level 'z' in the display list.
 * The window will be placed *on top* of any other
 * window with the same z-level.
 *
 * Returns the actual z-level you got,
 * or -1 if an error occurred.
 */
static int insert_window(ce_window_t *win, int z)
{
	if(z > _ce_windows)
		z = _ce_windows;
	if(_ce_windows + 1 > _ce_max_windows)
	{
		int nm = _ce_max_windows + 16;
		ce_window_t **nwt = (ce_window_t **)realloc(_ce_wintab,
				sizeof(void *) * nm);
		if(!nwt)
			return -1;
		_ce_wintab = nwt;
		_ce_max_windows = nm;
	}
	if(z < _ce_windows)
		memmove(_ce_wintab + z, _ce_wintab + z + 1,
				sizeof(void *) * (_ce_windows - z));
	++_ce_windows;
	_ce_wintab[z] = win;
	return z;
}


/*
 * Removes the window at z-level 'z' from the display list.
 */
static void remove_window_z(int z)
{
	if(z >= _ce_windows)
		return;
	if(z < _ce_windows-1)
		memmove(_ce_wintab + z + 1, _ce_wintab + z,
				sizeof(void *) * (_ce_windows - z - 1));
	--_ce_windows;
}


/*
 * Removes window 'win' from the display list.
 */
static void remove_window(ce_window_t *win)
{
	int z;
	for(z = 0; z < _ce_windows; ++z)
		if(_ce_wintab[z] == win)
		{
			remove_window_z(z);
			break;
		}
}


static void refresh(ce_window_t *win)
{
	ce_rect_t r;
	r.x = 0;
	r.y = 0;
	r.w = _ce_screen_w;
	r.h = _ce_screen_h;
	ce_get_screen(r, _ce_tempbuf, r.w);
	ce_put_screen(r, _ce_tempbuf, r.w);
//	ce_get_screen(win->view, _ce_tempbuf, win->view.w);
//	ce_put_screen(win->view, _ce_tempbuf, win->view.w);
}


/*----------------------------------------------------------
	Open/Close
----------------------------------------------------------*/

#define _ALL_CLASSES	(CE_OPEN_GRAPHICS | CE_OPEN_TEXT | CE_OPEN_STDIO)

int ce_open(int w, int h, int open_flags)
{
	int i;
	ce_rect_t r;
	int res = -1;

	if(!(open_flags & _ALL_CLASSES))
		open_flags |= _ALL_CLASSES;

	if(open_flags & CE_OPEN_GRAPHICS)
		res = _ce_sdl.open(w, h, open_flags);

	if((res < 0) && (open_flags & CE_OPEN_TEXT))
	{
		fprintf(stderr, "conemu: No \"real\" text targets"
				" supported yet!\n");
		res = -1;
	}

	if((res < 0) && (open_flags & CE_OPEN_STDIO))
		res = _ce_stdio.open(w, h, open_flags);

	w = _ce_screen_w;
	h = _ce_screen_h;

	_ce_tempbuf = malloc(w * h * sizeof(ce_character_t));
	if(!_ce_tempbuf)
	{
		fprintf(stderr, "conemu: Failed to allocate internal buffers!\n");
		ce_close();
		return -1;
	}

	load_palette();

	for(i = 0; i < w*h; ++i)
	{
		_ce_tempbuf[i].c = '?';
		_ce_tempbuf[i].fg = CE_LIGHTGRAY;
		_ce_tempbuf[i].bg = CE_BLACK;
		_ce_tempbuf[i].a = 0;
	}

	_ce_root_window = ce_open_window(w, h);
	if(!_ce_root_window)
	{
		fprintf(stderr, "conemu: Failed to open root window!\n");
		free(_ce_tempbuf);
		_ce_tempbuf = NULL;
		ce_close();
		return -1;
	}
	r.x = 0;
	r.y = 0;
	r.w = w;
	r.h = h;
	ce_move_window(_ce_root_window, &r);

	/* All done! */
	return 0;
}


void ce_close(void)
{
	if(_ce_root_window)
		ce_close_window(_ce_root_window);
	_ce_backend->close();
	_ce_backend = NULL;
	free(_ce_tempbuf);
	_ce_tempbuf = NULL;
}


/*----------------------------------------------------------
	Window management
----------------------------------------------------------*/

ce_window_t *ce_root_window(void)
{
	return _ce_root_window;
}


ce_window_t *ce_open_window(int w, int h)
{
	int i;
	ce_window_t *win = calloc(1, sizeof(ce_window_t));
	if(!win)
		return NULL;

	/* Hide */
	win->rect.x = win->rect.y = -1;
	win->rect.w = win->rect.h = 0;

	/* Set virtual size */
	win->virt.w = w;
	win->virt.h = h;

	/* Cursor to home */
	win->cx = win->cy = 0;

	/* Default colors */
	win->cursor_attribs.c = ' ';
	win->cursor_attribs.fg = CE_LIGHTGRAY;
	win->cursor_attribs.bg = CE_BLACK;

	/* Allocate buffer */
	win->buffer = malloc(w * h * sizeof(ce_character_t));
	if(!win->buffer)
	{
		free(win);
		return NULL;
	}
	for(i = 0; i < w*h; ++i)
		win->buffer[i] = win->cursor_attribs;

	insert_window(win, 0);

	return win;
}


int ce_scroll_window(ce_window_t *window, int x, int y)
{
	if((x == window->virt.x) && (y == window->virt.y))
		return 0;

	window->virt.x = x;
	window->virt.y = y;
	refresh(window);
	return 0;
}


int ce_move_window(ce_window_t *window, ce_rect_t *dest)
{
	ce_rect_t tmp = window->view;

	/* Hide window */
	memset(&window->view, 0, sizeof(window->view));

	/* Restore background */
	ce_get_screen(tmp, _ce_tempbuf, tmp.w);
	ce_put_screen(tmp, _ce_tempbuf, tmp.w);

	/* Calculate new window rect */
	if(CE_UNCHANGED == dest->x)
		dest->x = window->rect.x;
	if(CE_UNCHANGED == dest->y)
		dest->y = window->rect.y;
	if(CE_UNCHANGED == dest->w)
		dest->w = window->rect.w;
	if(CE_UNCHANGED == dest->h)
		dest->h = window->rect.h;
	window->rect = *dest;

	/* Calculate view (clipped) rect */
	if(dest->x < 0)
	{
		dest->w += dest->x;
		dest->x = 0;
	}
	if(dest->w > _ce_screen_w - dest->x)
	{
		if(_ce_screen_w > dest->x)
			dest->w = _ce_screen_w - dest->x;
		else
			dest->w = dest->h = 0;
	}
	if(dest->w)
	{
		if(dest->y < 0)
		{
			dest->h += dest->y;
			dest->y = 0;
		}
		if(dest->h > _ce_screen_h - dest->y)
		{
			if(_ce_screen_h > dest->y)
				dest->h = _ce_screen_h - dest->y;
			else
				dest->w = dest->h = 0;
		}
	}
	window->view = *dest;

	if(window->view.w && window->view.h)
	{
		refresh(window);
		return 1;
	}
	else
		return 0;
}


void ce_close_window(ce_window_t *window)
{
	ce_rect_t r;
	r.x = r.y = -1;
	r.w = r.h = 0;
	ce_move_window(window, &r);
	remove_window(window);
	free(window->buffer);
}


/*----------------------------------------------------------
	Visible cursor management
----------------------------------------------------------*/

int ce_cursor_at(ce_window_t *window, int x, int y)
{
	_ce_cursor_window = window;
	_ce_cursor_x = window->rect.x + x;
	_ce_cursor_y = window->rect.y + y;
	if(_ce_backend->cursor_at(_ce_cursor_x, _ce_cursor_y) < 0)
		return -1;	/* Backend error! --> */
	if(x < window->virt.x)
		return 0;
	if(y < window->virt.y)
		return 0;
	if(x + window->virt.x >= window->virt.w)
		return 0;
	if(y + window->virt.y >= window->virt.h)
		return 0;
	return 1;
}

void ce_cursor_off(void)
{
	ce_cursor_at(ce_root_window(), -1, -1);
}

ce_window_t *ce_find_cusor(void)
{
	return _ce_cursor_window;
}


/*----------------------------------------------------------
	Output cursor management
----------------------------------------------------------*/

int ce_locate(ce_window_t *window, int x, int y)
{
	int clip = 0;
	switch(x)
	{
	  case CE_UNCHANGED:
		x = window->cx;
		break;
	  case CE_INC:
		x = window->cx + 1;
		break;
	  case CE_DEC:
		x = window->cx - 1;
		break;
	}
	if(x < 0)
	{
		x = 0;
		clip = 1;
	}
	else if(x >= window->virt.w)
	{
		x = window->virt.w - 1;
		clip = 1;
	}

	switch(y)
	{
	  case CE_UNCHANGED:
		y = window->cy;
		break;
	  case CE_INC:
		y = window->cy + 1;
		break;
	  case CE_DEC:
		y = window->cy - 1;
		break;
	}
	if(y < 0)
	{
		y = 0;
		clip = 1;
	}
	else if(y >= window->virt.h)
	{
		y = window->virt.h - 1;
		clip = 1;
	}

	window->cx = x;
	window->cy = y;
	ce_cursor_at(window, x, y);

	if(_ce_backend->locate)
		_ce_backend->locate(window, x, y);

	return clip;
}

int ce_where_x(ce_window_t *window)
{
	if(_ce_backend->where_x)
		window->cx = _ce_backend->where_x(window);
	return window->cx;
}

int ce_where_y(ce_window_t *window)
{
	if(_ce_backend->where_y)
		window->cy = _ce_backend->where_y(window);
	return window->cy;
}

void ce_fgcolor(ce_window_t *window, unsigned char cl)
{
	window->cursor_attribs.fg = cl;
}

void ce_bgcolor(ce_window_t *window, unsigned char cl)
{
	window->cursor_attribs.bg = cl;
}


/*----------------------------------------------------------
	Window buffer management
----------------------------------------------------------*/

int ce_scroll_buffer(ce_window_t *window, int dx, int dy)
{
	int w, h, pitch;
	int y, i, j, yo, xot, xos, ydir;
	if(!dx && !dy)
		return 0;

	pitch = window->virt.w;
	w = window->virt.w - labs(dx);
	h = window->virt.h - labs(dy);
	xot = dx < 0 ? -dx : 0;
	xos = dx > 0 ? dx : 0;

	/* Scroll + clear left/right margin */
	if(dy <= 0)
	{
		y = 0;
		yo = 0;
		ydir = 1;
	}
	else
	{
		y = window->virt.h - 1;
		yo = dy;
		ydir = -1;
	}
	for(i = 0; i < h; ++i)
	{
		ce_character_t *line = window->buffer + y * pitch;
		memmove(line + xot, line + xos - pitch * dy,
				w * sizeof(ce_character_t));
		if(xot)
			for(j = 0; j < xot; ++j)
				line[j] = window->cursor_attribs;
		else if(xos)
			for(j = w; j < pitch; ++j)
				line[j] = window->cursor_attribs;
		y += ydir;
	}

	/* Clear top/bottom margin */
	if(dy)
	{
		if(dy < 0)
			y = window->virt.h + dy;
		else
			y = 0;
		h = labs(dy);
		for(i = 0; i < h; ++i)
		{
			ce_character_t *line = window->buffer +
					(i + y) * pitch;
			for(j = 0; j < pitch; ++j)
				line[j] = window->cursor_attribs;
		}
	}

	refresh(window);
	return 0;
}


void ce_clear(ce_window_t *window)
{
	int i;
	for(i = 0; i < window->virt.h * window->virt.w; ++i)
		window->buffer[i] = window->cursor_attribs;
	refresh(window);
	window->cx = window->cy = 0;
}


/*----------------------------------------------------------
	Low level access
----------------------------------------------------------*/

int ce_put(ce_window_t *window, int x, int y, ce_character_t c)
{
	/* Clip to virtual buffer */
	if(x < 0 || x >= window->virt.w)
		return 0;
	if(y < 0 || y >= window->virt.h)
		return 0;

	/* Write to virtual buffer */
	window->buffer[x + y * window->virt.w] = c;

	/* Calculate screen coords */
	x -= window->virt.x;
	y -= window->virt.y;
	x += window->rect.x;
	y += window->rect.y;

	/* Clip to viewport */
	if(x < window->view.x || x >= window->view.x + window->view.w)
		return 1;
	if(y < window->view.y || y >= window->view.y + window->view.h)
		return 1;

	/* Write to screen */
	if(_ce_backend->put)
		return _ce_backend->put(x, y, c);
	else
	{
		ce_rect_t r;
		r.x = x;
		r.y = y;
		r.w = 1;
		r.h = 1;
		_ce_backend->put_fast(r, &c, 1);
		return 1;
	}
}


ce_character_t ce_get(ce_window_t *window, int x, int y)
{
	/* Clip to virtual buffer */
	if(x < 0 || x >= window->virt.w)
		return _ce_blank;
	if(y < 0 || y >= window->virt.h)
		return _ce_blank;

	/* Read from virtual buffer */
	return window->buffer[x + y * window->virt.w];
}


int ce_put_rect(ce_window_t *window, ce_rect_t rect,
		ce_character_t *buffer, int pitch)
{
	int i;

	/* Clip to virtual buffer */
	if(rect.x + rect.w < 0 || rect.x >= window->virt.w)
		return 0;
	if(rect.y + rect.h < 0 || rect.y >= window->virt.h)
		return 0;
	if(rect.x < 0)
	{
		buffer -= rect.x;
		rect.w += rect.x;
		rect.x = 0;
	}
	if(rect.w > window->virt.w - rect.x)
		rect.w = window->virt.w - rect.x;
	if(rect.y < 0)
	{
		buffer -= rect.y * pitch;
		rect.h += rect.y;
		rect.y = 0;
	}
	if(rect.h > window->virt.h - rect.y)
		rect.h = window->virt.h - rect.y;

	/* Write to virtual buffer */
	for(i = rect.y; i < rect.y + rect.h; ++i)
		memcpy(window->buffer + rect.x + i * window->virt.w,
				buffer + i * pitch, rect.w);

	/* Calculate screen coords */
	rect.x -= window->virt.x;
	rect.y -= window->virt.y;
	rect.x += window->rect.x;
	rect.y += window->rect.y;

	/* Clip to viewport */
	if(rect.x + rect.w < window->view.x || rect.x >= window->view.w)
		return 0;
	if(rect.y + rect.h < window->view.y || rect.y >= window->view.h)
		return 0;
	if(rect.x < window->view.x)
	{
		buffer += window->view.x - rect.x;
		rect.w -= window->view.x - rect.x;
		rect.x = window->view.x;
	}
	if(rect.w > window->view.w - rect.x)
		rect.w = window->view.w - rect.x;
	if(rect.y < window->view.y)
	{
		buffer += (window->view.y - rect.y) * pitch;
		rect.h -= window->view.y - rect.y;
		rect.y = window->view.y;
	}
	if(rect.h > window->view.h - rect.y)
		rect.h = window->view.h - rect.y;

	_ce_backend->put_fast(rect, buffer, pitch);

	return 1;
}


int ce_get_rect(ce_window_t *window, ce_rect_t rect,
		ce_character_t *buffer, int pitch)
{
	int i;

	/* Clip to virtual buffer */
	if(rect.x + rect.w < 0 || rect.x >= window->virt.w)
		return 0;
	if(rect.y + rect.h < 0 || rect.y >= window->virt.h)
		return 0;
	if(rect.x < 0)
	{
		buffer -= rect.x;
		rect.w += rect.x;
		rect.x = 0;
	}
	if(rect.w > window->virt.w - rect.x)
		rect.w = window->virt.w - rect.x;
	if(rect.y < 0)
	{
		buffer -= rect.y * pitch;
		rect.h += rect.y;
		rect.y = 0;
	}
	if(rect.h > window->virt.h - rect.y)
		rect.h = window->virt.h - rect.y;

	/* Copy from virtual buffer */
	for(i = 0; i < rect.h; ++i)
		memcpy(buffer + i * pitch,
				window->buffer + rect.x +
				(i + rect.y) * window->virt.w,
				rect.w * sizeof(ce_character_t));
	return 1;
}


/*
 * Like ce_(get|put)_rect(), but with the following
 * differences:
 *
 *	* THEY DO NOT PERFORM CLIPPING! Bad or Weird
 *	  things may happen if you copy to/from areas
 *	  outside the screen.
 *
 *	* ce_put_screen() works directly towards the
 *	  screen or underlying API where applicable.
 *
 *	* ce_put_screen() does not affect or depend
 *	  upon the contents of any windows. (In fact,
 *	  it totally ignores the existence of windows!)
 *
 *	* ce_put_screen() does not care about window
 *	  positions, overlap and the like. It just
 *	  brutally overwrites anything in the area
 *	  of the "screen" specified by 'rect'.
 *
 *	* ce_put_screen() is *usually* faster than
 *	  ce_put_rect().
 *
 *	* ce_get_screen() is *slower* than
 *	  ce_get_rect() on most targets. (It's
 *	  actually just an internal tool used on some
 *	  targets, to assemble parts of windows into
 *	  a buffer, for refreshing the "screen".)
 *
 * Returns 0 upon success, or -1 if an error occured.
 */
int ce_put_screen(ce_rect_t rect, ce_character_t *buffer, int pitch)
{
	return _ce_backend->put_fast(rect, buffer, pitch);
}


int ce_get_screen(ce_rect_t rect, ce_character_t *buffer, int pitch)
{
	ce_rect_t r;
	int i;
	/*
	 * Quick hack brute force implementetion: Just
	 * render all windows into the buffer, using
	 * painter's algo. :-)
	 */
	for(i = _ce_windows - 1; i >= 0; --i)
	{
		int Amin, Amax, Bmin, Bmax;
		ce_window_t *w = _ce_wintab[i];
		int bx, by;

		/* Where to read from in the virtual buffer */
		r.x = w->virt.x;
		r.y = w->virt.y;
		r.w = w->view.w;
		r.h = w->view.h;

		/*
		 * Clip to within the specified rect.
		 * (stolen from SDL_surface.c)
		 */

		/* Where to write to, in the buffer */
		bx = w->view.x;
		by = w->view.y;

		/* Horizontal intersection */
		Amin = r.x;
		Amax = Amin + r.w;
		Bmin = rect.x;
		Bmax = Bmin + rect.w;
		if(Bmin > Amin)
		        Amin = Bmin;
		else
			bx += Amin - Bmin;
		r.x = Amin;
		if(Bmax < Amax)
		        Amax = Bmax;
		if(Amax - Amin > 0)
			r.w = Amax - Amin;
		else
			continue;

		/* Vertical intersection */
		Amin = r.y;
		Amax = Amin + r.h;
		Bmin = rect.y;
		Bmax = Bmin + rect.h;
		if(Bmin > Amin)
		        Amin = Bmin;
		else
			by += Amin - Bmin;
		r.y = Amin;
		if(Bmax < Amax)
		        Amax = Bmax;
		if(Amax - Amin > 0)
			r.h = Amax - Amin;
		else
			continue;

		/* Read from window! */
		if(ce_get_rect(w, r, buffer + bx + by*pitch, pitch) < 0)
			return -1;
	}
	return 0;
}


void ce_set_palette(unsigned char index,
		unsigned char r, unsigned char g, unsigned char b)
{
	if(_ce_backend->set_palette)
		_ce_backend->set_palette(index, r, g, b);
}


/*----------------------------------------------------------
	Basic input helpers
----------------------------------------------------------*/

/*
 * Returns 1 if a key is available from the keyboard,
 * 0 if not, or -1 if keyboard polling is not possible
 * on this target.
 */
int ce_kbhit(void)
{
	if(_ce_backend->kbhit)
		return _ce_backend->kbhit();
	else
		return -1;
}

/*
 * This ain't the way it's supposed to be...
 * But right now, the only *real* backend is SDL,
 * so we can't really do without it anyway.
 */
#include "SDL.h"
void ce_delay(int ms)
{
	SDL_Delay(ms);
}


/*----------------------------------------------------------
	stdio style input
----------------------------------------------------------*/

static int _ce_putback = 0;
static int _ce_putback_char;

int ce_getc(void)
{
	if(_ce_putback)
	{
		_ce_putback = 0;
		return _ce_putback_char;
	}
	else
		return _ce_backend->getc();
}

int ce_ungetc(int c)
{
	if(!_ce_putback)
	{
		_ce_putback_char = c;
		_ce_putback = 1;
		return c;
	}
	return EOF;
}

int ce_scanf(ce_window_t *window, const char *fmt, ...)
{
	int res;
	va_list args;
	if(!_ce_backend->scanf)
	{
		fprintf(stderr, "conemu: Default scanf() not implemented!\n");
		return -1;
	}
	va_start(args, fmt);
	res = _ce_backend->scanf(window, fmt, args);
	va_end(args);
	return res;
}


/*----------------------------------------------------------
	stdio style calls (the fast way)
----------------------------------------------------------*/

int ce_putchar(ce_window_t *window, int c)
{
	ce_character_t temp;
	if(_ce_backend->putchar)
		return _ce_backend->putchar(window, c);

	switch(c)
	{
	  case CE_BS:
		if(!window->cx)
			return EOF;	/* Hit margin! */
		--window->cx;
		ce_putchar(window, ' ');	/* Oooh, recursive! :-) */
		--window->cx;
		break;
	  case CE_TAB:
		/* A TAB *is* 8 characters. Period! :-D */
		if(!(window->cx & 0x7))
			ce_putchar(window, ' ');
		if(0 == window->cx)
			break;
		while(window->cx & 0x7)
		{
			ce_putchar(window, ' ');
			if(0 == window->cx)
				break;
		}
		break;
	  case CE_LF:
		window->cx = 0;
		++window->cy;
		if(window->cy >= window->virt.h)
		{
			window->cy = window->virt.h - 1;
			ce_scroll_buffer(window, 0, -1);
		}
		break;
	  case CE_CR:
		window->cx = 0;
		break;
	  case CE_UP:
		ce_locate(window, CE_UNCHANGED, CE_DEC);
		break;
	  case CE_DOWN:
		ce_locate(window, CE_UNCHANGED, CE_INC);
		break;
	  case CE_LEFT:
		ce_locate(window, CE_DEC, CE_UNCHANGED);
		break;
	  case CE_RIGHT:
		ce_locate(window, CE_INC, CE_UNCHANGED);
		break;
	  case CE_DEL:
		/* Why do anything stupid and useless here? */
		break;
	  default:
		if(c < 32)
			break;

		temp = window->cursor_attribs;
		temp.c = c;
		ce_put(window, window->cx, window->cy, temp);
		++window->cx;
		if(window->cx >= window->virt.w)
		{
			window->cx = 0;
			++window->cy;
			if(window->cy >= window->virt.h)
			{
				window->cy = window->virt.h - 1;
				ce_scroll_buffer(window, 0, -1);
			}
		}
		break;
	}

	ce_cursor_at(window, window->cx, window->cy);
	return c & 0xff;
}


int ce_puts(ce_window_t *window, const char *s)
{
	int count = 0;
	if(_ce_backend->puts)
		return _ce_backend->puts(window, s);

	while(*s)
	{
		if(ce_putchar(window, *s++) < 0)
			return EOF;
		++count;
	}
	return count;
}


int ce_printf(ce_window_t *window, const char *fmt, ...)
{
	int res;
	va_list args;
	va_start(args, fmt);
	if(_ce_backend->printf)
	{
		res = _ce_backend->printf(window, fmt, args);
		va_end(args);
		return res;
	}

/////////////////////////////////////

res = ce_puts(window, fmt);

/////////////////////////////////////

	va_end(args);
	return res;
}


/*----------------------------------------------------------
	stdio redirection (the easy way)
----------------------------------------------------------*/

#if 0
int ce_redirect_stdio(ce_window_t *window);
#endif
