/*
------------------------------------------------------------
	conemu.h - Text Console Emulator / Wrapper
------------------------------------------------------------
 * Copyright (C) David Olofson, 2002
 *
 * This code is released under the terms of the GNU LGPL.
 */

/*
 * Note that all coordinate systems have (0, 0) as the
 * top-left corner, *not* (1, 1) or whatever, as many (?)
 * other text mode APIs!
 */

#ifndef _CONEMU_H_
#define _CONEMU_H_

/*----------------------------------------------------------
	Data types and structures
----------------------------------------------------------*/

/* Your basic rectangle, for passing coords and size around. */
typedef struct ce_rect_t
{
	signed short	x, y;
	unsigned short	w, h;
} ce_rect_t;

/*
 * A single character + colors + attribute flags.
 */
typedef struct ce_character_t
{
	unsigned char	c;
	unsigned char	fg, bg;
	unsigned char	a;
} ce_character_t;

/*
 * This struct holds the characters of a window,
 * with colors and attributes. The buffer can be
 * bigger than the visible, on-screen window.
 */
typedef struct ce_window_t
{
	/* Position and size */
	ce_rect_t	rect;

	/* Viewport (rect clipped to the "screen") */
	ce_rect_t	view;

	/* Virtual size and scroll position */
	ce_rect_t	virt;

	/* The buffer */
	ce_character_t	*buffer;

	/*
	 * Cursor position.
	 * (Note that the actual visible cursor
	 * may be in another window!)
	 */
	int		cx, cy;

	/* Cursor colors and attributes */
	ce_character_t	cursor_attribs;
} ce_window_t;

/* Handy enum for the built-in standard colors */
typedef enum ce_colors_t
{
	CE_BLACK = 0,
	CE_BLUE,
	CE_GREEN,
	CE_CYAN,
	CE_RED,
	CE_MAGENTA,
	CE_BROWN,
	CE_LIGHTGRAY,

	CE_DARKGRAY,
	CE_LIGHTBLUE,
	CE_LIGHTGREEN,
	CE_LIGHTCYAN,
	CE_LIGHTRED,
	CE_LIGHTMAGENTA,
	CE_YELLOW,
	CE_WHITE
} ce_colors_t;


/* conemu "ASCII" control codes */
typedef enum ce_codes_t
{
	CE_NULL = 0,

	CE_BS = 8,	/* Backspace */
	CE_TAB = 9,
	CE_LF = 10,	/* Line Feed */

	CE_CR = 13,	/* Carriage Return */

	CE_RIGHT = 16,	/* Cursor Right */
	CE_LEFT = 17,	/* Cursor Left */

	CE_ESC = 27,

	CE_UP = 30,	/* Cursor Up */
	CE_DOWN = 31,	/* Cursor Down */

	CE_DEL = 127	/* Delete */
} ce_codes_t;


/*----------------------------------------------------------
	Defines
----------------------------------------------------------*/

/*
 * Give conemu a hint as to what kind of console emulation
 * you want.
 *
 * If more than one of _TEXT, _GRAPHICS and _STDIO are
 * specified, they will be tried in that order, until
 * conemu manages to set up one of them.
 *
 * If you specify CE_OPEN_ANYWHERE, *all* supported
 * targets will be tried, "most powerful" ones first,
 * whereas if you explicitly specify one or more target
 * classes, only the targets in those classes will be
 * tried - and if all fail, ce_open() fails. Thus, you
 * can try target classes in any order you like, using
 * multiple calls to ce_open(), specifying only one
 * class at a time.
 */

/* Don't care... Give me *anything*! */
#define CE_OPEN_ANYWHERE	0x00000000

/*
 * Try "real" text targets, such as Un*x console, Windows
 * "DOS box", or raw video text mode.
 */
#define CE_OPEN_TEXT		0x00000001

/* Try emulation over graphics, such as the SDL target. */
#define CE_OPEN_GRAPHICS	0x00000002

/*
 * Try using conemu as a simple stdio wrapper.
 * (Most advanced features are either disabled,
 * or potentially very slow.)
 */
#define CE_OPEN_STDIO		0x00000004

/* Try to open the "screen" in fullscreen mode. */
#define CE_OPEN_FULLSCREEN	0x00000010


/*----------------------------------------------------------
	Open/Close
----------------------------------------------------------*/

/*
 * Open the conemu "screen". The 'w' and 'h' arguments are
 * in characters (not pixels!), and like the 'open_flags'
 * argument, they're just a *hint*. You may not always get
 * exactly what you ask for!
 *
 * ce_open() will also create the root window, which is
 * initially of the same size as the "screen". The only
 * differences between the root window and other windows
 * are:
 *	* The root window is enlarged as needed, to
 *	  guarantee that it's always at least the size
 *	  of the screen.
 *
 *	* The root window is automatically resized
 *	  to fit the "screen", if the "screen" is
 *	  resized.
 *
 *	* The root window is visible by default.
 *
 *	* The root window should *not* be freed by
 *	  applications - it belongs to the screen.
 *
 * Returns 0 on success, or -1 in case of failure.
 */
int ce_open(int w, int h, int open_flags);

/*
 * Close the conemu "screen". All conemu windows and other
 * objects will be deleted. All comemu calls except for
 * ce_open() will have undefined behavior after this call.
 * (They will most likely just segfault. :-)
 */
void ce_close(void);


/*----------------------------------------------------------
	Window management
----------------------------------------------------------*/

#define	CE_INC		32765
#define	CE_DEC		32766
#define	CE_UNCHANGED	32767

/*
 * Returns the address of the root window struct.
 */
ce_window_t *ce_root_window(void);

/*
 * Create a new conemu window of the specified virtual
 * size. The window will initially be invisible (which
 * is done by locating it at (-1, -1) and setting it's
 * size to 1x1), so you have to call ce_move_window()
 * to make it visible.
 *
 * Returns the new window, or NULL, if an error occured.
 */
ce_window_t *ce_open_window(int w, int h);

/*
 * Change the virtual scroll position of 'window' to
 * ('x', 'y'). The change will instantly take effect
 * on the screen.
 *
 * The coordinate value CE_UNCHANGED can be passed
 * for 'x', 'y' or both to preserve the old values.
 * Passing CE_UNCHANGED for both 'x' and 'y' will
 * simply return the result of the "is window wrapped"
 * test.
 *
 * Returns 1 if the scroll operation resulted in the
 * window wrapping (ie some areas outside the window's
 * virtual buffer would be visible on-screen), 0
 * otherwise, or -1 if an error occured.
 */
int ce_scroll_window(ce_window_t *window, int x, int y);

/*
 * Move and resize the on-screen view of 'window' so
 * that it occupies the area of the "screen" described
 * by 'dest'. 'dest' will be overwritten with the
 * resulting position and size of the window.
 *
 * The coordinate value CE_UNCHANGED can be passed
 * for one or more of the fields of 'dest', to preserve
 * the old values of the respective fields. Passing
 * CE_UNCHANGED for all fields will simply return the
 * result of the window visibility test, and the window
 * position and size.
 *
 * Returns 0 if the window became invisible, 1 if the
 * window is partially visible, 2 if the entire window
 * (that is, the area specified by 'dest'; not the full
 * virtual buffer) is visible, or -1 if an error occured.
 */
int ce_move_window(ce_window_t *window, ce_rect_t *dest);

/*
 * Closes 'window', freeing up all of it's resources.
 */
void ce_close_window(ce_window_t *window);


/*----------------------------------------------------------
	Visible cursor management
----------------------------------------------------------*/

/*
 * Place the visible cursor at virtual position ('x', 'y')
 * inside 'window'. Setting a position outside the visible
 * part of the window, or outside the virtual window
 * disables the cursor.
 *
 * This function DOES NOT change the position of the output
 * cursor of the window.
 *
 * Returns 1 if the cursor is visible at the new position,
 * 0 if not, or -1 if an error occured.
 */
int ce_cursor_at(ce_window_t *window, int x, int y);

/*
 * Turns the visible cursor off. Equivalent to:
 *	ce_cursor_at(ce_root_window(), -1, -1);
 */
void ce_cursor_off(void);

/*
 * Return the window that currently owns the visible cursor.
 */
ce_window_t *ce_find_cusor(void);


/*----------------------------------------------------------
	Output cursor management
----------------------------------------------------------*/

/*
 * These should be fairly obvious - but keep in mind
 * that the top-left corner is still (0, 0), and that
 * it's the top-left corner of the *virtual buffer*,
 * not the screen, not the visible part of the window,
 * or anything else. :-)
 *
 * ce_locate() accepts CE_UNCHANGED for either of the
 * x or y arguments, to preserve the current value.
 * It also supports the codes CE_INC and CE_DEC, which
 * increment respectively decrement the current value
 * by one.
 *
 * ce_locate() returns 1 if the cursor, as a result of
 * the operation, had to be restricted to within the
 * buffer. Otherwise, it returns 0.
 */
int ce_locate(ce_window_t *window, int x, int y);
int ce_where_x(ce_window_t *window);
int ce_where_y(ce_window_t *window);
void ce_fgcolor(ce_window_t *window, unsigned char cl);
void ce_bgcolor(ce_window_t *window, unsigned char cl);


/*----------------------------------------------------------
	Window buffer management
----------------------------------------------------------*/

/*
 * Scroll the contents of 'window's virtual buffer down
 * by 'dy' lines, and right by 'dx' lines. Areas that
 * "scroll in" from outside the buffer are filled with
 * the contents of the 'ce_clear_character' global
 * variable.
 *
 * Returns 0 upon success, or -1 if an error occured.
 */
int ce_scroll_buffer(ce_window_t *window, int dx, int dy);

/*
 * Clear the window buffer and return cursor to the
 * top-left corner.
 */
void ce_clear(ce_window_t *window);


/*----------------------------------------------------------
	Low level access
----------------------------------------------------------*/

/*
 * Put character 'c' at ('x', 'y') in 'window'. The
 * operation writes to the virtual off-screen buffer
 * of the text window, and also causes the actual display
 * to be instantly updated.
 *
 * Note that 'c' is *NOT* a C "char", but a conemu character
 * struct, containing colors and attribute flags!
 *
 * Returns 1 if the character ended up inside the virtual
 * screen buffer, 0 if not, or -1 if an error occured.
 */
int ce_put(ce_window_t *window, int x, int y, ce_character_t c);

/*
 * Get the character at ('x', 'y') from 'window'.
 *
 * Note that return is by *value*; not by reference.
 *
 * If ('x', 'y') is outside the virtual buffer of the
 * window, a blank character with black background and
 * foreground colors, and attribute flags cleared, is
 * returned.
 */
ce_character_t ce_get(ce_window_t *window, int x, int y);

/*
 * Copy characters, colors and attributes from 'buffer'
 * into the area of 'window' specified by 'rect'. The
 * operation writes into the virtual off-screen buffer
 * of 'window', and also causes the actual display to be
 * instantly updated.
 *
 * 'pitch' is the number of ce_character_t elements per
 * row of 'buffer'. Normally, you should pass 'rect.w',
 * but you can use it to support buffers larger than the
 * areas copied.
 *
 * Note that the buffer must be an array of ce_character_t,
 * *NOT* C "char", and that the buffer must be at least
 * (rect.w * rect.h * sizeof(ce_character_t)) bytes big -
 * or Bad Things may happen!
 *
 * Returns 1 if the operation affected any currently
 * visible characters, 0 if not, or -1 if an error occured.
 */
int ce_put_rect(ce_window_t *window, ce_rect_t rect,
		ce_character_t *buffer, int pitch);

/*
 * Copies the area of 'window' specified by 'rect' into
 * 'buffer'. Copying is done from the virtual buffer of
 * the window. Characters outside the virtual buffer of
 * 'window' are left unchanged in 'buffer'.
 *
 * 'pitch' is the number of ce_character_t elements per
 * row of 'buffer'. Normally, you should pass 'rect.w',
 * but you can use it to support buffers larger than the
 * areas copied.
 *
 * Note that the buffer must be an array of ce_character_t,
 * *NOT* C "char", and that the buffer must fit at least
 * (rect.w * rect.h * sizeof(ce_character_t)) bytes - or
 * Bad Things *will* happen!
 *
 * Returns 1 if anything was copied, 0 if not (ie 'rect'
 * was entirely outside 'window'), or -1 if an error
 * occured.
 */
int ce_get_rect(ce_window_t *window, ce_rect_t rect,
		ce_character_t *buffer, int pitch);

/*
 * Like ce_(get|put)_rect(), but with the following
 * differences:
 *
 *	* THEY DO NOT PERFORM CLIPPING! Bad or Weird
 *	  things may happen if you copy to/from areas
 *	  outside the screen.
 *
 *	* ce_put_screen() works directly towards the
 *	  screen or underlying API where applicable.
 *
 *	* ce_put_screen() does not affect or depend
 *	  upon the contents of any windows. (In fact,
 *	  it totally ignores the existence of windows!)
 *
 *	* ce_put_screen() does not care about window
 *	  positions, overlap and the like. It just
 *	  brutally overwrites anything in the area
 *	  of the "screen" specified by 'rect'.
 *
 *	* ce_put_screen() is *usually* faster than
 *	  ce_put_rect().
 *
 *	* ce_get_screen() is *slower* than
 *	  ce_get_rect() on most targets. (It's
 *	  actually just an internal tool used on some
 *	  targets, to assemble parts of windows into
 *	  a buffer, for refreshing the "screen".)
 *
 * Returns 0 upon success, or -1 if an error occured.
 */
int ce_put_screen(ce_rect_t rect, ce_character_t *buffer, int pitch);
int ce_get_screen(ce_rect_t rect, ce_character_t *buffer, int pitch);


/*
 * Set palette entry 'index' to the best match of
 * the color defined by 'r', 'g' and 'b'.
 *
 * Note that not all targets will be able to give
 * you the exact color. Some targets have only 16
 * fixed colors to chose from, and some may have
 * even less.
 */
void ce_set_palette(unsigned char index,
		unsigned char r, unsigned char g, unsigned char b);


/*----------------------------------------------------------
	Basic input helpers
----------------------------------------------------------*/

/*
 * Returns 1 if a key is available from the keyboard,
 * 0 if not, or -1 if keyboard polling is not possible
 * on this target.
 */
int ce_kbhit(void);

/*
 * Wait for 'ms' milliseconds, and then returns.
 */
void ce_delay(int ms);


/*----------------------------------------------------------
	stdio style input
----------------------------------------------------------*/

int ce_getc(void);
int ce_ungetc(int c);
int ce_scanf(ce_window_t *window, const char *fmt, ...);


/*----------------------------------------------------------
	stdio style calls (the fast way)
----------------------------------------------------------*/

int ce_putchar(ce_window_t *window, int c);
int ce_puts(ce_window_t *window, const char *s);
int ce_printf(ce_window_t *window, const char *fmt, ...);


/*----------------------------------------------------------
	stdio redirection (the easy way)
----------------------------------------------------------*/

int ce_redirect_stdio(ce_window_t *window);

#endif /* _CONEMU_H_ */
